/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann t t-online.de
 Begin of project:			05.06.2021
 Latest version generated:	21.09.2021
 Filename:					StepMotor.c
 Description:				Control of Left and Right Step Motor Functions via controller device L297 / L298
 Microprocessor:			ATmega8

 Berechnung von TIMER0 und TIMER2 Counter Start Wert mit Mhroboter.xlsx

 ********************************************************/
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>

#include "General.h"
#include "RS232.h"
#include "Delay.h"
#include "StepMotor.h"
/*
** Define here the step motor mode, either full or half step mode.
*/
#define FULL_STEP		// Note: Pin 19 of IC L297 "H/F" of step motor driver shall go to GND
//#define HALF_STEP		// Note: Pin 19 of IC L297 "H/F" of step motor driver shall go to +5V
/*
** Defines for right step motor
*/
#define RIGHT_CW_CCW_DDR	SBIT(DDRD, 2)
#define RIGHT_CW_CCW_OUT	SBIT(PORTD, 2)
#define RIGHT_CLOCK_DDR		SBIT(DDRC, 3)
#define RIGHT_CLOCK_OUT		SBIT(PORTC, 3)
#define RIGHT_RESET_DDR		SBIT(DDRD, 4)
#define RIGHT_RESET_OUT		SBIT(PORTD, 4)
#define RIGHT_ENABLE_DDR	SBIT(DDRC, 2)
#define RIGHT_ENABLE_OUT	SBIT(PORTC, 2)
/*
** Defines for left step motor
*/
#define LEFT_CW_CCW_DDR		SBIT(DDRD, 6)
#define LEFT_CW_CCW_OUT		SBIT(PORTD, 6)
#define LEFT_CLOCK_DDR		SBIT(DDRD, 5)
#define LEFT_CLOCK_OUT		SBIT(PORTD, 5)
#define LEFT_RESET_DDR		SBIT(DDRB, 1)
#define LEFT_RESET_OUT		SBIT(PORTB, 1)
#define LEFT_ENABLE_DDR		SBIT(DDRC, 0)
#define LEFT_ENABLE_OUT		SBIT(PORTC, 0)

#define	MOTOR_CLOCK_PERIOD_LOW	1			// 1 us, see L297 datasheet, which specifies min = 0.5 us

/*
** Private Variables.
*/
volatile uint8_t	iRunningLeft;			// Flag indicating Left motor running
volatile uint8_t	iLimitStepsFlagLeft;	// Element of enum SM_STEPS
volatile uint16_t	iMaxStepsLeft;			// Maximum steps to go in case of SM_STEPS_LIMITED

volatile uint8_t	iRunningRight;			// Flag indicating Right motor running
volatile uint8_t	iLimitStepsFlagRight;	// Element of enum SM_STEPS
volatile uint16_t	iMaxStepsRight;			// Maximum steps to go in case of SM_STEPS_LIMITED

volatile uint8_t	StartCountValueLeft;	// Start value for TIMER0 for left step motor
volatile uint8_t	StartCountValueRight;	// Start value for TIMER2 for right step motor

volatile uint8_t	SlowSpeedFactorLeft;		// The factor for a reduced speed used in ISR (TIMER0_OVF_vect)
volatile uint8_t	SlowSpeedFactorCounterLeft;	// The down counter for the slow speed factor used in ISR (TIMER0_OVF_vect)

volatile uint8_t	SlowSpeedFactorRight;			// The factor for a reduced speed used in ISR (TIMER2_OVF_vect)
volatile uint8_t	SlowSpeedFactorCounterRight;	// The down counter for the slow speed factor used in ISR (TIMER2_OVF_vect)

/*******************************************************
 Public Function: SM_Init

 Purpose: Initialise the left and right step motor controller L297.

 Input parameters: None

 Return value: void
 *******************************************************/
void SM_Init (void)
	{
/*
** Set left and right step motor ports to output.
*/
	LEFT_CW_CCW_DDR = 1;
	LEFT_CLOCK_DDR = 1;
	LEFT_RESET_DDR = 1;
	LEFT_ENABLE_DDR = 1;

	RIGHT_CW_CCW_DDR = 1;
	RIGHT_CLOCK_DDR = 1;
	RIGHT_RESET_DDR = 1;
	RIGHT_ENABLE_DDR = 1;
/*
** Set left and right step motor control pins of L297 to default configuration.
*/
	LEFT_CW_CCW_OUT = SM_DIR_CW;		// Set turn direction to ClockWise.
	LEFT_ENABLE_OUT = 0;				// Disable motor (no current)
	LEFT_CLOCK_OUT = 1;					// Pulse low to do a step
	LEFT_RESET_OUT = 1;					// Pulse low to reset

	RIGHT_CW_CCW_OUT = SM_DIR_CCW;		// Set turn direction to Counter ClockWise.
	RIGHT_ENABLE_OUT = 0;				// Disable motor (no current)
	RIGHT_CLOCK_OUT = 1;				// Pulse low to do a step
	RIGHT_RESET_OUT = 1;				// Pulse low to reset
/*
** Set slow speed factor to default.
*/
	SlowSpeedFactorLeft = 1;
	SlowSpeedFactorCounterLeft = SlowSpeedFactorLeft;
	SlowSpeedFactorRight = 1;
	SlowSpeedFactorCounterRight = SlowSpeedFactorLeft;
/*
** Reset left and right motor controller L297.
*/
	Delay_us (500);
	SM_Left_Reset ();
	SM_Right_Reset ();
/*
** Indicate, that motors are not running.
*/
	iRunningLeft = FALSE;
	iRunningRight = FALSE;
/*
** Set limit of steps to go.
*/
	iLimitStepsFlagLeft = SM_STEPS_UNLIMITED;
	iLimitStepsFlagRight = SM_STEPS_UNLIMITED;
/*
** Set TIMER0 (left motor) and TIMER2 (right motor) start count value to default value.
** Note: These 8-bit timers generate an overflow interrupt.
*/
	StartCountValueLeft = 0;
	StartCountValueRight = 0;
/*
** Configure TIMER0 (left Motor) and TIMER2 (right Motor) interrupts.
** Motor is:
**	- Type: PSM35BYG104 (Pollin), bought 2016
**	- 1.8 degree / Step	--> 200 Steps for one turn in full step mode
**	- In addition we have a gear with ratio of 3.3333, hence 667 steps for one turn (full step mode)
**
** For the left step motor:
** ------------------------
** Set TIMER0 to mode "Overflow" with prescaler (be careful: is different as for TIMER2):
**	CS02 = 1, CS01 = 0, CS00 = 1	--> Prescaler = 1024 for FULL_STEP mode
**	CS02 = 1, CS01 = 0, CS00 = 0	--> Prescaler = 256  for HALF_STEP mode
*/
#ifdef FULL_STEP
	TCCR0 = (1<<CS02) | (1<<CS00);		// Prescaler = 1024
#endif
#ifdef HALF_STEP
	TCCR0 = (1<<CS02);					// Prescaler = 256
#endif
	TIMSK |= (1<<TOIE0);				// Set Timer/Counter0 Overflow Interrupt Enable
/*
** For the right step motor:
** -------------------------
** Set TIMER2 to mode "Overflow" with prescaler (be careful: is different as for TIMER0):
**	CS22 = 1, CS21 = 1, CS20 = 1	--> Prescaler = 1024 for FULL_STEP mode
**	CS22 = 1, CS21 = 1, CS20 = 0	--> Prescaler = 256  for HALF_STEP mode
**	WGM21 = 0, WGM20 = 0			--> Normal Counter Up Mode
*/
#ifdef FULL_STEP
	TCCR2 = (1<<CS22) | (1<<CS21) | (1<<CS20);		// Prescaler = 1024
#endif
#ifdef HALF_STEP
	TCCR2 = (1<<CS22) | (1<<CS21);					// Prescaler = 256
#endif
	TIMSK |= (1<<TOIE2);							// Set Timer/Counter2 Overflow Interrupt Enable
/*
** Enable global interrupts.
*/
	sei ();

	return;
	}
/*******************************************************
 Public Function: SM_Left_Start

 Purpose: Start Left step motor.

  Input parameters:
	uint8_t		iDirection		{SM_LEFT_AHEAD, SM_LEFT_BACK}
	uint8_t		iSpeed			{2, ... , 100} in percent of max speed
	uint16_t	iNumSteps		{0, ... , 65.535}	0 = unlimited

 Return value:	void
 *******************************************************/
void SM_Left_Start (uint8_t iDirection, uint8_t iSpeed, uint16_t iNumSteps)
	{
/*
** Disable interrupts, because we do not want to be disturbed here.
*/
	cli ();
/*
** Set the speed and if required, the maximum number of motor steps.
*/
	SM_Left_SetSpeed (iSpeed);
	iLimitStepsFlagLeft = SM_STEPS_UNLIMITED;
	if (iNumSteps > 0)
		{
		iLimitStepsFlagLeft = SM_STEPS_LIMITED;
		iMaxStepsLeft = iNumSteps;
		}
/*
** Reset L297 of motor, set turn direction and enable motor.
*/
	SM_Left_Reset ();
	LEFT_CW_CCW_OUT = iDirection;
	LEFT_ENABLE_OUT = 1;
	iRunningLeft = TRUE;
/*
** Enable interrupts.
*/
	sei ();

	return;
	}
/*******************************************************
 Public Function: SM_Right_Start

 Purpose: Start Right step motor.

  Input parameters:
	uint8_t		iDirection		{SM_RIGHT_AHEAD, SM_RIGHT_BACK}
	uint8_t		iSpeed			{2, ... , 100} in percent of max speed
	uint16_t	iNumSteps		{0, ... , 65.535}	0 = unlimited

 Return value:	void
 *******************************************************/
void SM_Right_Start (uint8_t iDirection, uint8_t iSpeed, uint16_t iNumSteps)
	{
/*
** Disable interrupts, because we do not want to be disturbed here.
*/
	cli ();
/*
** Set the speed and if required, the maximum number of motor steps.
*/
	SM_Right_SetSpeed (iSpeed);
	iLimitStepsFlagRight = SM_STEPS_UNLIMITED;
	if (iNumSteps > 0)
		{
		iLimitStepsFlagRight = SM_STEPS_LIMITED;
		iMaxStepsRight = iNumSteps;
		}
/*
** Reset L297 of motor, set turn direction and enable motor.
*/
	SM_Right_Reset ();
	RIGHT_CW_CCW_OUT = iDirection;
	RIGHT_ENABLE_OUT = 1;
	iRunningRight = TRUE;
/*
** Enable interrupts.
*/
	sei ();

	return;
	}
/*******************************************************
 Public Function: SM_Left_Stop

 Purpose: Stop the Left motor.

 Input parameters: None

 Return value: void
 *******************************************************/
void SM_Left_Stop (void)
	{
/*
** Disable L297 of Left motor.
*/
	LEFT_ENABLE_OUT = 0;
	iRunningLeft = FALSE;

	return;
	}
/*******************************************************
 Public Function: SM_Right_Stop

 Purpose: Stop the Right motor.

 Input parameters: None

 Return value: void
 *******************************************************/
void SM_Right_Stop (void)
	{
/*
** Disable L297 of Right motor.
*/
	RIGHT_ENABLE_OUT = 0;
	iRunningRight = FALSE;

	return;
	}
/*******************************************************
 Public Function: SM_IsRightRunning

 Purpose: Return whether Right Motor is running

 Input parameters: None

 Return value:	uint8_t
	TRUE	Motor is still running
	FALSE	Motor is not running
 *******************************************************/
uint8_t SM_IsRightRunning (void)
	{
	return iRunningRight;
	}
/*******************************************************
 Public Function: SM_IsLeftRunning

 Purpose: Return whether Left Motor is running

 Input parameters: None

 Return value:	uint8_t
	TRUE	Motor is still running
	FALSE	Motor is not running
 *******************************************************/
uint8_t SM_IsLeftRunning (void)
	{
	return iRunningLeft;
	}
/*******************************************************
 Public Function: SM_Left_Step

 Purpose: Perform ONE left Motor Step

 Input Parameter: None

 Return Value: void
*******************************************************/
void SM_Left_Step (void)
	{
/*
** Set clock to low level, then wait for MOTOR_CLOCK_PERIOD_LOW period and then set back to high level.
** Datasheet states 0.5 us minimum.
*/
	LEFT_CLOCK_OUT = 0;
	Delay_us (MOTOR_CLOCK_PERIOD_LOW);
	LEFT_CLOCK_OUT = 1;

	return;
	}
/*******************************************************
 Public Function: SM_Right_Step

 Purpose: Perform ONE right Motor Step

 Input Parameter: None

 Return Value: void
*******************************************************/
void SM_Right_Step (void)
	{
/*
** Set clock to low level, then wait for MOTOR_CLOCK_PERIOD_LOW period and then set back to high level.
** Datasheet states 0.5 us minimum.
*/
	RIGHT_CLOCK_OUT = 0;
	Delay_us (MOTOR_CLOCK_PERIOD_LOW);
	RIGHT_CLOCK_OUT = 1;

	return;
	}
/*******************************************************
 Public Function: SM_Left_Reset

 Purpose: Reset Left Motor Electronics

 Input Parameter: None

 Return Value: void
*******************************************************/
void SM_Left_Reset (void)
	{
	iRunningLeft = FALSE;

	LEFT_RESET_OUT = 0;
	Delay_us (2);			// minimum 1 us low level, see datasheet
	LEFT_RESET_OUT = 1;

	return;
	}
/*******************************************************
 Public Function: SM_Right_Reset

 Purpose: Reset Right Motor Electronics

 Input Parameter: None

 Return Value: void
*******************************************************/
void SM_Right_Reset (void)
	{
	iRunningRight = FALSE;

	RIGHT_RESET_OUT = 0;
	Delay_us (2);			// minimum 1 us low level, see datasheet
	RIGHT_RESET_OUT = 1;

	return;
	}
/*******************************************************
 Public Function: SM_Left_SetSpeed

 Purpose: Set Speed of Left Motor Electronics based on TIMER0

 Input Parameter:
	uint8_t		iSpeed = {2, ... , 100} in percent of max speed

 Return Value: void
*******************************************************/
void SM_Left_SetSpeed (uint8_t iSpeed)
	{
/*
** Compute the value for the Counter Register of TIMER0.
** Formula: y = 1.5 * x + 89 = x + 0.5x + 89
** Set the start value for the counter.
** Note: Counter counts up to 256 and then generates an overflow interrupt.
** See ISR (TIMER0_OVF_vect).
*/
	StartCountValueLeft = iSpeed + (iSpeed >> 1) + 89;

	return;
	}
/*******************************************************
 Public Function: SM_Right_SetSpeed

 Purpose: Set Speed of Right Motor Electronics based on TIMER2

 Input Parameter:
	uint8_t		iSpeed = {2, ... , 100} in percent of max speed

 Return Value: void
*******************************************************/
void SM_Right_SetSpeed (uint8_t iSpeed)
	{
/*
** Compute the value for the Counter Register of TIMER2.
** Formula: y = 1.5 * x + 89 = x + 0.5x + 89
** Set the start value for the counter.
** Note: Counter counts up to 256 and then generates an overflow interrupt.
** See ISR (TIMER2_OVF_vect).
*/
	StartCountValueRight = iSpeed + (iSpeed >> 1) + 89;

	return;
	}
/*******************************************************
 Public Function: SM_SetLeftSlowSpeedFactor

 Purpose: Set the slow speed factor:
			iFactor = 1 is nominal speed
			iFactor = MAX_SLOW_SPEED_FACTOR gives MAX_SLOW_SPEED_FACTOR times reduced speed

 Input Parameter:
	uint8_t		iFactor = {1, ... , MAX_SLOW_SPEED_FACTOR}
				In case of a wrong input value, the factor will be set to 1

 Return Value: void
*******************************************************/
void SM_SetLeftSlowSpeedFactor (uint8_t iFactor)
	{
	SlowSpeedFactorLeft = iFactor;
	SlowSpeedFactorCounterLeft = SlowSpeedFactorLeft;
	return;
	}
/*******************************************************
 Public Function: SM_SetRightSlowSpeedFactor

 Purpose: Set the slow speed factor:
			iFactor = 1 is nominal speed
			iFactor = MAX_SLOW_SPEED_FACTOR gives MAX_SLOW_SPEED_FACTOR times reduced speed

 Input Parameter:
	uint8_t		iFactor = {1, ... , MAX_SLOW_SPEED_FACTOR}

 Return Value: void
*******************************************************/
void SM_SetRightSlowSpeedFactor (uint8_t iFactor)
	{
	SlowSpeedFactorRight = iFactor;
	SlowSpeedFactorCounterRight = SlowSpeedFactorRight;
	return;
	}
/*******************************************************
 Public Function: SM_GetLeftSlowSpeedFactor

 Purpose: Get the slow speed factor:

 Input Parameter: void

 Return Value: uint8_t
*******************************************************/
uint8_t SM_GetLeftSlowSpeedFactor (void)
	{
	return SlowSpeedFactorLeft;
	}
/*******************************************************
 Public Function: SM_GetRightSlowSpeedFactor

 Purpose: Get the slow speed factor

 Input Parameter: void

 Return Value: uint8_t
*******************************************************/
uint8_t SM_GetRightSlowSpeedFactor (void)
	{
	return SlowSpeedFactorRight;
	}
/*******************************************************
 TIMER0 Overflow Interrupt for Left Motor

*******************************************************/
ISR (TIMER0_OVF_vect)
	{
	uint8_t		tmp_sreg;
/*
** Save status register.
*/
	tmp_sreg = SREG;
/*
** Update the start value of the TIMER0 counter.
*/
	TCCR0 = StartCountValueLeft;
/*
** Consider slow speed factor.
** Do a left motor step, if required.
** In case we have to do limited steps, do the associated procedure.
*/
	if (iRunningLeft)
		{
		if (SlowSpeedFactorLeft != 1)
			{
			SlowSpeedFactorCounterLeft--;
			if (SlowSpeedFactorCounterLeft != 0) goto ret;
			SlowSpeedFactorCounterLeft = SlowSpeedFactorLeft;
			}

		SM_Left_Step ();
		if (iLimitStepsFlagLeft == SM_STEPS_LIMITED)
			{
			iMaxStepsLeft--;
			if (iMaxStepsLeft == 0) SM_Left_Stop ();	// will set iRunningLeft to FALSE
			}
		}
/*
** Restore status register, hence also the I-Flag of the status register.
*/
ret:
	SREG = tmp_sreg;
	return;
	}
/*******************************************************
 TIMER2 Overflow Interrupt for Right Motor

*******************************************************/
ISR (TIMER2_OVF_vect)
	{
	uint8_t		tmp_sreg;
/*
** Save status register.
*/
	tmp_sreg = SREG;
/*
** Update the start value of the TIMER2 counter.
*/
	TCCR2 = StartCountValueRight;
/*
** Consider slow speed factor.
** Do a right motor step, if required.
** In case we have to do limited steps, do the associated procedure.
*/
	if (iRunningRight)
		{
		if (SlowSpeedFactorRight != 1)
			{
			SlowSpeedFactorCounterRight--;
			if (SlowSpeedFactorCounterRight != 0) goto ret;
			SlowSpeedFactorCounterRight = SlowSpeedFactorRight;
			}

		SM_Right_Step ();
		if (iLimitStepsFlagRight == SM_STEPS_LIMITED)
			{
			iMaxStepsRight--;
			if (iMaxStepsRight == 0) SM_Right_Stop ();	// will set iRunningRight to FALSE
			}
		}
/*
** Restore status register, hence also the I-Flag of the status register.
*/
ret:
	SREG = tmp_sreg;
	return;
	}
