/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann t t-online.de
 Begin of project:			18.07.2021
 Latest version generated:	22.09.2021
 Filename:					main.c
 Description:				Main routine for Step Motor Slave Controller
 Microprocessor:			ATmega8

 User Timers:
	TIMER0 for left step motor
	TIMER2 for right step motor

 Used Interrupts:
	ISR (TIMER0_OVF_vect)	in StepMotor.c	Do Left Motor steps
	ISR (TIMER2_OVF_vect)	in StepMotor.c	Do Right Motor steps
 ********************************************************/
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <string.h>

#include "General.h"
#include "RS232.h"
#include "Delay.h"
#include "TWI_Slave.h"
#include "StepMotor.h"
/*
** Function Declaration.
*/
void GetAndDoCommand ();
void ConfirmCommand (uint8_t iValue);
/*
** Definition of TWI commands.
** This enum list shall be identical with the one in the Master controller software !
*/
enum
	{
	COMMAND_SM_PING,					// Send a ping command to the TWI slave controller
	COMMAND_SM_BOOT,					// Boot the Step Motor Slave Controller
	COMMAND_SM_SET_SPEED,				// Set the motor speed
	COMMAND_SM_START,					// Start the motor
	COMMAND_SM_STOP,					// Stop the motor
	COMMAND_SM_RESET,					// Reset the motor
	COMMAND_SM_SET_SLOW_SPEED_FACOR,	// Set slow speed factor
	COMMAND_SM_IS_RUNNING,				// Are motors still running ?
	COMMAND_SM_CONFIRM,					// Confirm the command
	};
/*
** Global variables.
*/
uint8_t			SW_Version = 1;
TWI_SM_STRUCT	SM_Buf;
/********************************************************
  Main Function
*********************************************************/
int main (void)
	{
/*
** Clear any interrupt.
*/
	cli ();
	Delay_ms (50);
/*
** Init RS-232 interface.
*/
	RS232_Init ();
	printf ("StepMotor-Controller\n");
	printf ("Version %i: %s %s\n", SW_Version, __DATE__, __TIME__);
/*
** Init TWI Interface with TWI address = 2
*/
	TWIS_Init (2);
/*
** Init the Step Motor functionality.
*/
	SM_Init ();
/*
** Main Loop starts: Check on commands from TWI
*/
	while (TRUE)
		{
		GetAndDoCommand ();
		}
	}
/*******************************************************
 Public Function: GetAndDoCommand

 Purpose: Get and do a command, as received from the TWI master

 The following commands will be processed by the slave:

	COMMAND_SM_PING,					// Send a ping command to the TWI slave controller
	COMMAND_SM_BOOT,					// Boot the Step Motor Slave Controller
	COMMAND_SM_SET_SPEED,				// Set the motor speed
	COMMAND_SM_START,					// Start the motor
	COMMAND_SM_STOP,					// Stop the motor
	COMMAND_SM_RESET,					// Reset the motor
	COMMAND_SM_SET_SLOW_SPEED_FACOR,	// Set slow speed factor
	COMMAND_SM_IS_RUNNING,				// Are motors still running ?

 This enum list shall be identical with the one in the Master controller software !
 The command shall return the first 2 bytes in SM_Buf, which are:
	- SM_Buf.iCommand
	- SM_Buf.iMotor
 Each received command has to be confirmed by placing COMMAND_SM_CONFIRM to SM_Buf.iCommand.
 For the command COMMAND_SM_IS_RUNNING, the status (TRUE or FALSE) shall be returned in SM_Buf.iMotor.

 Input parameters: void

 Return value: void
 *******************************************************/
void GetAndDoCommand (void)
	{
	uint8_t		iNumBytes;
	uint16_t	iValue16;
/*
** Check, whether we have received via the TWI interface a command.
** If Yes, then perform the command.
**
** To speed up the "if (SM_Buf.iCommand == xxx)" instructions, we sort them with respect to
** their probability of occurrence. Highest probability first.
*/
	if (TWIS_ReadDataFromTWI (&iNumBytes, (uint8_t*)&SM_Buf) == TRUE)
		{
/*
** Command: COMMAND_SM_IS_RUNNING
** ==============================
** If the parameter in SM_Buf.iMotor is SM_ONE_MOTOR, then:
** - We return in ConfirmCommand() the value TRUE, when at least one motor is still running.
** - If none of the two motors are running, we return FALSE in ConfirmCommand().
*/
		if (SM_Buf.iCommand == COMMAND_SM_IS_RUNNING)
			{
			//printf ("Is running ?\n");
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) ConfirmCommand (SM_IsLeftRunning ());
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) ConfirmCommand (SM_IsRightRunning ());
			if (SM_Buf.iMotor == SM_ONE_MOTOR) ConfirmCommand (SM_IsLeftRunning () | SM_IsRightRunning ());
			return;
			}
/*
** Command: COMMAND_SM_START
** =========================
*/
		if (SM_Buf.iCommand == COMMAND_SM_START)
			{
			//printf ("Start\n");
			iValue16 = (SM_Buf.iPara3 << 8) + SM_Buf.iPara4; // Create a 16 bit word out of iPara3 and iPara4
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) SM_Left_Start (SM_Buf.iPara1, SM_Buf.iPara2, iValue16);
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) SM_Right_Start (SM_Buf.iPara1, SM_Buf.iPara2, iValue16);
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_STOP
** ========================
*/
		if (SM_Buf.iCommand == COMMAND_SM_STOP)
			{
			//printf ("Stop\n");
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) SM_Left_Stop ();
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) SM_Right_Stop ();
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_SET_SPEED
** =============================
*/
		if (SM_Buf.iCommand == COMMAND_SM_SET_SPEED)
			{
			//printf ("Set Speed\n");
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) SM_Left_SetSpeed (SM_Buf.iPara1);
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) SM_Right_SetSpeed (SM_Buf.iPara1);
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_RESET
** =========================
*/
		if (SM_Buf.iCommand == COMMAND_SM_RESET)
			{
			//printf ("Reset\n");
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) SM_Left_Reset ();
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) SM_Right_Reset ();
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_SET_SLOW_SPEED_FACTOR
** =========================================
*/
		if (SM_Buf.iCommand == COMMAND_SM_SET_SLOW_SPEED_FACOR)
			{
			//printf ("Set Slow Speed\n");
			if (SM_Buf.iMotor == SM_LEFT_MOTOR) SM_SetLeftSlowSpeedFactor (SM_Buf.iPara1);
			if (SM_Buf.iMotor == SM_RIGHT_MOTOR) SM_SetRightSlowSpeedFactor (SM_Buf.iPara1);
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_PING
** ========================
*/
		if (SM_Buf.iCommand == COMMAND_SM_PING)
			{
			ConfirmCommand (0);
			return;
			}
/*
** Command: COMMAND_SM_BOOT
** ========================
** Confirm the command and then start the watch dog, which will boot the controller after 15 ms.
*/
		if (SM_Buf.iCommand == COMMAND_SM_BOOT)
			{
			//printf ("Boot\n");
			ConfirmCommand (0);
			cli ();
			CLEAR_BIT (GICR, INT0);
			wdt_enable (WDTO_15MS);
			while (TRUE) {};
			}
/*
** We ignore here any other commands, which may come.
** ==================================================
*/
		//printf ("C=Wrong\n");
		ConfirmCommand (0);
		}

	return;
	}
/*******************************************************
 Public Function: ConfirmCommand

 Purpose: Confirm the received TWI command
			Note: From the struct SM_Buf we only return to the master the two first bytes, which are:
				uint8_t SM_Buf.iCommand;	// The command byte
				uint8_t SM_Buf.iMotor;		// The motor to be addressed {SM_LEFT, SM_RIGHT};

 Input parameters:
	uint8_t		iValue		The value to transmit back to the master in SM_Buf.iMotor

 Return value: void
 *******************************************************/
void ConfirmCommand (uint8_t iValue)
	{
	SM_Buf.iCommand = COMMAND_SM_CONFIRM;
	SM_Buf.iMotor = iValue;
	TWIS_WriteDataToTWI (2, (uint8_t*)&SM_Buf);
	return;
	}